"""
This module contains classes and functions for running automated experiments and also for 
optimizing hyperparameters using custom methods and other methods like the optuna module for 
optimizing hyperparameters.
"""

import logging
import optuna
from lxfx.utils import createLogger, handleExit
from lxfx.xfx import XFX

import json
import os

class XFXExperiment():
    def __init__(self):
        pass
    
    def run(self):
        pass

class XFXOptunaExperiment():
    """
    This class uses Optuna to tune hyperparameters such as learning rate, number of layers, layer hidden sizes, optimizer, etc.
    To flag a parameter as tunable by Optuna, replace its value in the config file or config dict with a dict:
    {
        "tune": true,
        "data_type": data_type,
        "value_range": [val_1, val_2]
    }
    If a parameter is not flagged as tunable, its default value will be used.
    """
    def __init__(self, n_trials = 10,
                 n_epochs = 10,
                 config_file_path = None, 
                 config_dict = None, 
                 model_type = None):
        
        self.console_logger = createLogger(is_consoleLogger=True, log_level=logging.INFO)

        self.study = optuna.create_study(direction = "minimize")
        self.n_trials = n_trials
        self.n_epochs = n_epochs

        self.config_file_path = config_file_path 
        self.config_dict = config_dict
        self.load_config_dict()

        self.model_type = model_type 
        if self.model_type is None:
            self.console_logger("Model type not set. Cannot continue ...")
            handleExit()


        self.tunable_parameters = {}
        self.get_tunable_parameters()

        self.xfx_project = None

    def load_config_dict(self):
        try:
            if not self.config_dict:
                if self.config_file_path and os.path.exists(self.config_file_path):
                    with open(self.config_file_path, 'r') as file:
                        self.config_dict = json.load(file)
                else:
                    self.console_logger.error(f"File: {self.config_file_path} not found")
                    handleExit()
        except json.JSONDecodeError as e:
            self.console_logger.error(f"JSON decode error: {str(e)}")
            handleExit()
        except FileNotFoundError as e:
            self.console_logger.error(f"File not found: {str(e)}")
            handleExit()
        except Exception as e:
            self.console_logger.error(f"An unexpected error occurred while loading the config dict: {str(e)}")
            handleExit()
        if self.config_dict is None:
            self.console_logger.error("Config dict is empty. Cannot continue ...")
            handleExit()

    def get_param_dict(self, value, section, param):
        try:
            param_dict = {
                "section": section,
                "data_type": value.get("data_type"),
                "value_range": value.get("value_range"),
                "idx": None,  # Default value for idx
                "nested":False
            }
            if param_dict["data_type"] is None or param_dict["value_range"] is None:
                raise ValueError(f"Parameter {param} in section {section} is missing 'data_type' or 'value_range'.")
            return param_dict
        except Exception as e:
            self.console_logger.error(f"Error in get_param_dict for parameter {param}: {str(e)}")
            handleExit()

    def get_tunable_parameters(self):
        try:
            for section, section_dict in self.config_dict.items():
                if not isinstance(section_dict, dict):
                    self.console_logger.error(f"Section {section} is not a dictionary.")
                    continue
                for param, value in section_dict.items():
                    if isinstance(value, dict) and value.get("tune"):
                        self.tunable_parameters[param] = self.get_param_dict(value, section, param)
                    elif isinstance(value, list):
                        for idx, val in enumerate(value):
                            if isinstance(val, dict) and val.get("tune"):
                                param_dict = self.get_param_dict(val, section, param)
                                param_dict["idx"] = idx
                                self.tunable_parameters[param] = param_dict
                            # this is the maximum number of nested lists the config can have
                            elif isinstance(val, list):
                                for nested_idx, nested_val in enumerate(val):
                                    if isinstance(nested_val, dict) and nested_val.get("tune"):
                                        param_dict = self.get_param_dict(nested_val, section, param)
                                        param_dict["idx"] = nested_idx
                                        param_dict["nested"] = True
                                        param_dict["parent_idx"] = idx  # Index of the nested list in the first list
                                        self.tunable_parameters[param] = param_dict

        except Exception as e:
            self.console_logger.error(f"An error occurred while getting tunable parameters: {str(e)}")
            handleExit()

    def objective(self, trial: optuna.Trial):
        try:
            for param, param_dict in self.tunable_parameters.items():
                if param_dict["data_type"] == "float":
                    param_dict["value"] = trial.suggest_float(param, param_dict["value_range"][0], param_dict["value_range"][1])
                elif param_dict["data_type"] == "int":
                    param_dict["value"] = trial.suggest_int(param, param_dict["value_range"][0], param_dict["value_range"][1])
                elif param_dict["data_type"] == "categorical":
                    param_dict["value"] = trial.suggest_categorical(param, param_dict["value_range"])
                
                # change the value in the config dict
                if param_dict["idx"] is None:
                    self.config_dict[param_dict["section"]][param] = param_dict["value"]
                elif param_dict["nested"] is True:
                    self.config_dict[param_dict["section"]][param][param_dict["parent_idx"]][param_dict["idx"]] = param_dict["value"]
                else:
                    self.config_dict[param_dict["section"]][param][param_dict["idx"]] = param_dict["value"]

            # Open an XFX project for each trial
            self.xfx_project = XFX(_config_dict=self.config_dict, model_type=self.model_type)
            return self.xfx_project.train(epochs=self.n_epochs, tuna_exp=True)
        except Exception as e:
            self.console_logger.error(f"An error occurred in the objective function: {str(e)}")
            handleExit()

    def run(self):
        try:
            self.get_tunable_parameters()
            self.study.optimize(self.objective, n_trials=self.n_trials)
        except Exception as e:
            self.console_logger.error(f"An error occurred during the optimization run: {str(e)}")
            handleExit()


import logging
import json
from icecream import IceCreamDebugger
import os
import atexit

from networkx import non_edges

from lxfx.utils import createLogger, handleExit
from lxfx.data import TimeSeriesDatasetManager
from lxfx.trainers import ModelTrainer
from lxfx.predPlotter import PredPlotter
from lxfx.models import FxFCModel, FxFCAutoEncoder

class XFX:
    def __init__(self, config_file_path = None, _config_dict = None, 
                 model_type = None, manual_seed = None, disable_ic_debug = True):
        
        # setup debugging mechanisms
        self.disable_ic_debug = disable_ic_debug
        self.console_logger = createLogger(is_consoleLogger=True)
        self.ic_debugger = IceCreamDebugger(prefix="XFXDebug: |")
        if self.disable_ic_debug:
            self.ic_debugger.disable()
    
        self.config_file_path = config_file_path
        self._config_dict = _config_dict
        self.model_type = model_type
        self.dataset_manager = None 

        self.dataset_config = None
        self.trainer_config = None
        self.FxFCModel_config = None
        self.FxFCAutoEncoder_config = None

        self.trainer = None 
        self.dataset_manager = None
        self.model = None
        self.manual_seed = manual_seed
        
        self.model_num_features = None
        self.model_out_features = None
        self.model_future_pred_length = None

        self.initializeXFXProject()

    def initializeXFXProject(self):
        self.get_config()
        self.dataset_manager = self.get_dataset_manager()
        if self.manual_seed:
            self.dataset_manager.set_seed(self.manual_seed)
        self.model = self.get_model()
        self.trainer = self.get_trainer()
        atexit.register(self.save_config)

    def get_config(self):
        if self.config_file_path:
            try:
                self.config = json.load(open(self.config_file_path))
            except Exception as e:
                self.console_logger.error(f"Error loading config file: {e}")
                handleExit()
        else:
            self.config = self._config_dict

        # get datasetManager and trainer 
        try:
            self.dataset_config = self.config["TimeSeriesDatasetManager"]
        except Exception as e:
            self.console_logger.error(f"Error loading dataset config: {e}")
            handleExit()
        try:
            self.trainer_config = self.config["ModelTrainer"]
        except Exception as e:
            self.console_logger.error(f"Error loading trainer config: {e}")
            handleExit()
        # FxFCModel config
        try:
            self.FxFCModel_config = self.config["FxFCModel"]
        except Exception as e:
            if self.model_type == "FxFCModel":
                self.console_logger.error(f"Error loading FxFCModel config: {e}")
                self.console_logger.error("Please check the config file for the FxFCModel section.")
                handleExit()
            else:
                self.FxFCModel_config = None

        try:
            self.FxFCAutoEncoder_config = self.config["FxFCAutoEncoder"]
        except Exception as e:
            if self.model_type == "FxFCAutoEncoder":
                self.console_logger.error(f"Error loading FxFCAutoEncoder config: {e}")
                self.console_logger.error("Please check the config file for the FxFCAutoEncoder section.")
                handleExit()
            else:
                self.FxFCAutoEncoder_config = None

    def get_dataset_manager(self):
        if self.dataset_manager is None:
            self.dataset_manager = TimeSeriesDatasetManager(
                data=None,
                file_path=self.dataset_config.get("file_path"),
                targets=None,
                future_pred_length=self.dataset_config.get("future_pred_length", 1),
                targetIndices=self.dataset_config.get("targetIndices", None),
                target_names=self.dataset_config.get("target_names"),
                drop_names=self.dataset_config.get("drop_names", None),
                date_col=self.dataset_config.get("date_col"),
                sequence_length=self.dataset_config.get("sequence_length", 5),
                stride=self.dataset_config.get("stride", 1),
                scaler_feature_range=self.dataset_config.get("scaler_feature_range", [0, 1]),
                transform=self.dataset_config.get("transform", None),
                use_default_scaler=self.dataset_config.get("use_default_scaler", True),
                split_criterion=self.dataset_config.get("split_criterion", [0.7, 0.15, 0.15]),
                batchsize=self.dataset_config.get("batchsize", 8),
                is_fx=self.dataset_config.get("is_fx", False),
                transform_targets=self.dataset_config.get("transform_targets", True),
                index_col=self.dataset_config.get("index_col", None),
                manual_seed=self.dataset_config.get("manual_seed", None),
                is_autoregressive=self.dataset_config.get("is_autoregressive", False),
                test_sequence_length=self.dataset_config.get("test_sequence_length", None), 
                test_stride=self.dataset_config.get("test_stride", None)
            )

        self.model_num_features = self.dataset_manager.model_num_features
        self.model_out_features = self.dataset_manager.model_out_features
        self.model_future_pred_length = self.dataset_manager.model_future_pred_length
        self.ic_debugger(self.model_num_features, self.model_out_features, self.model_future_pred_length)

        return self.dataset_manager

    def get_trainer(self):
        if self.trainer is None:
            self.trainer = ModelTrainer(
                model=self.model,
                dataset_manager=self.dataset_manager,
                loss_fn=self.trainer_config.get("loss_fn", "mse"),
                optimizer=self.trainer_config.get("optimizer", "adam"),
                epochs=self.trainer_config.get("epochs", 3),
                accuracy_fn=self.trainer_config.get("accuracy_fn", None),
                lr=self.trainer_config.get("lr", 0.001),
                project_path=self.trainer_config.get("project_path"),
                is_logging=self.trainer_config.get("is_logging", True),
                resume=self.trainer_config.get("resume", False),
            )
        return self.trainer

    def get_model(self): 
        if self.model is None:
            if self.model_type == "FxFCModel":
                self.model = FxFCModel(
                    num_features=self.model_num_features,
                    block_type=self.FxFCModel_config.get("block_type", "lstm"),
                    out_features=self.model_out_features,
                    units=self.FxFCModel_config.get("units"),
                    num_layers=self.FxFCModel_config.get("num_layers", 1),
                    out_units=self.FxFCModel_config.get("out_units"),
                    activation=self.FxFCModel_config.get("activation", "tanh"),
                    bidirectional=self.FxFCModel_config.get("bidirectional", False),
                    pass_states=self.FxFCModel_config.get("pass_states", False),
                    use_batch_norm=self.FxFCModel_config.get("use_batch_norm", False),
                    pass_block_hidden_state=self.FxFCModel_config.get("pass_block_hidden_state", False)
                )
            elif self.model_type == "FxFCAutoEncoder":
                self.model = FxFCAutoEncoder(
                    num_features=self.model_num_features,
                    target_features=self.model_out_features,
                    future_pred_length=self.model_future_pred_length,
                    block_types=self.FxFCAutoEncoder_config.get("block_types", ["lstm", "lstm"]),
                    units=self.FxFCAutoEncoder_config.get("units"),
                    out_units=self.FxFCAutoEncoder_config.get("out_units"),
                    num_layers=self.FxFCAutoEncoder_config.get("num_layers", [1, 1]),
                    activations=self.FxFCAutoEncoder_config.get("activations", ["tanh", "tanh"]),
                    latent_dim=self.FxFCAutoEncoder_config.get("latent_dim", 32),
                    dropout=self.FxFCAutoEncoder_config.get("dropout", [0.2, 0.2]),
                    bidirectional=self.FxFCAutoEncoder_config.get("bidirectional", [False, False]),
                    use_batch_norm=self.FxFCAutoEncoder_config.get("use_batch_norm", [False, False]),
                    pass_block_hidden_state=self.FxFCAutoEncoder_config.get("pass_block_hidden_state", [False, False]), 
                    is_attentive=self.FxFCAutoEncoder_config.get("is_attentive", False),
                    pass_states=self.FxFCAutoEncoder_config.get("pass_states", [False ,False])
                )
            else:
                self.console_logger.error(f"Invalid model type: {self.model_type}")
                handleExit()
        return self.model
        
    def train(self, epochs = None, tuna_exp = False):
        train_metrics = self.trainer.trainModel(epochs=epochs, tuna_exp = tuna_exp)
        self.save_config()
        if tuna_exp:
            return train_metrics # last validation loss

    def predict(self):
        self.trainer.predict()

    def save_config(self):
        config_dict = {}
        config_dict["model_type"] = self.model_type
        config_dict["model"] = self.model.config_dict()
        config_dict["dataset"] = self.dataset_manager.config_dict()
        config_dict["trainer"] = self.trainer.config_dict()

        project_dir = self.trainer.project_path
        config_file_path = os.path.join(project_dir, "config.json")
        with open(config_file_path, "w") as f:
            json.dump(config_dict, f, indent=4)
        print(f"Config file saved to: {config_file_path}")

    def plotPredictions(self, preds, figsize = (6,6), test_only=False,
                        _len = 100, test_plots = None):
        """
        Plots a graph(s) showing the predictions against the actual values.
        """
        predPlotter = PredPlotter(preds, self.dataset_manager, test_only=test_only,
                                  length = _len, figsize = figsize, test_plots = test_plots)
        predPlotter.plotPredictions()


import json
import os
import logging

from lxfx.utils import createLogger

class XFXConfig:
    """
    This class generates the configuration data for the XFX model.
    Configure the model that is going to be instantiated by the XFX class. This can be either of the following:
    1. FxFCModel
    2. FxFCEncoder
    3. FxFCDecoder
    4. FxFCAutoEncoder
    Specify the model type by the model_type parameter in the XFX class and the class shall
    automatically load the appropriate configuration data and model for that type.
    If the project directory specified is already an XFX project then it is loaded as a project with
    existing configurations, models and all information saved during the previous runs.
    """
    def __init__(self, dump_dir = None, config_file_path = None, config_file_name = None):
        self.console_logger = createLogger(is_consoleLogger=True, log_level=logging.ERROR)
        self.dump_dir = dump_dir
        self.config_file_path = config_file_path
        self.config_file_name = config_file_name
        self.supported_model_types = ["FCBlock", "FxFCModel", "FxFCEncoder", "FxFCDecoder", "FxFCAutoEncoder"]
        self.FCBlock_model_config = {
                "in_features": None,
                "hidden_size": None,
                "out_size": None,
                "nature": "lstm",
                "dropout": 0.2,
                "num_layers": 1,
                "bidirectional": False,
                "activation": "tanh",
                "use_batch_norm": False,
                "pass_block_hidden_state": False
        }
        self.FXFCAutoEncoder_model_config = {
                "future_pred_length": 1,
                "block_types": ["lstm", "lstm"],
                "comment0": "The current configuration requires that the hidden size of the last layer of the encoder and the first layer of the decoder are the same.",
                "units": [[], []],
                "comment1": "Once the out_units is set then pass_block_hidden_states must be false beucase the first layer shall have a hidden state whose shape is different from the next layer's hidden state shape",
                "out_units": [[], []],
                "num_layers": [1, 1],
                "activations": ["tanh", "tanh"],
                "latent_dim": 32,
                "dropout": [0.2, 0.2],
                "bidirectional": [False, False],
                "use_batch_norm": [False, False],
                "pass_states": [True, True],
                "pass_block_hidden_state": [False, False],
                "is_attentive": False
        }
        self.FXFCModel_model_config = {
                "block_type": "lstm",
                "units": None,
                "comment0": "The length of the units array is equal to the number or fxfc blocks",
                "num_layers": 1,
                "comment1": "The number of out_units MUST be equal to the number of units",
                "out_units": None,
                "comment2": "Once the out_units is set then pass_block_hidden_states must be false beucase the first layer shall have a hidden state whose shape is different from the next layer's hidden state shape",
                "activation": "tanh",
                "bidirectional": False,
                "pass_states": False,
                "comment": "use_batch_norm is a MUST if the data is forex data",
                "use_batch_norm": False,
                "pass_block_hidden_state": False
        }
        self.FXFCEncoder_model_config = {
                "block_type": None,
                "units": None,
                "out_units": None,
                "num_layers": 1,
                "activation": "tanh",
                "latent_dim": None,
                "use_batch_norm": False,
                "bidirectional": False
        }
        self.FXFCDecoder_model_config = {
                "latent_dim": None,
                "block_type": None,
                "units": None,
                "out_units": None,
                "num_layers": 1,
                "activation": "tanh",
                "use_batch_norm": False,
                "initialize_weights": False,
                "initializer_method": None
        }

        self.models_config = {
            "FCBlock": self.FCBlock_model_config,
            "FxFCModel": self.FXFCModel_model_config,
            "FxFCEncoder": self.FXFCEncoder_model_config,
            "FxFCDecoder": self.FXFCDecoder_model_config,
            "FxFCAutoEncoder": self.FXFCAutoEncoder_model_config,
        }

        self.config_dict = {
            "TimeSeriesDatasetManager": {
                "file_path": None,
                "future_pred_length": 1,
                "targetIndices": None,
                "target_names": [],
                "drop_names": None,
                "date_col": None,
                "sequence_length": 5,
                "test_sequence_length": 2,
                "is_autoregressive": False,
                "stride": 1,
                "scaler_feature_range": [0, 1],
                "transform": None,
                "use_default_scaler": True,
                "split_criterion": [0.7, 0.15, 0.15],
                "batchsize": None,
                "is_fx": False,
                "transform_targets": True,
                "index_col": None,
                "manual_seed": 42,
                "test_stride": None
            },
            "ModelTrainer": {
                "loss_fn": "mse",
                "optmizer": "adam",
                "epochs": 3,
                "accuracy_fn": None,
                "lr": 0.001,
                "project_path": None,
                "is_logging": True,
                "resume": False,
                "targets_transformed": False
            },
            "TensorInitializer": {
                "method": None
            }
        }

    def get_config_dict(self):
        return self.config_dict

    def dump_config_file(self, model_type = None): # Write the dictionary to a JSON file
        if model_type in self.supported_model_types:
            filepath = "config.json"
            if self.dump_dir:
                filepath = os.path.join(self.dump_dir, filepath)
            elif self.config_file_path:
                filepath = self.config_file_path
            elif self.config_file_name:
                filepath = os.path.join(os.getcwd(), self.config_file_name)

            self.config_dict[model_type] = self.models_config[model_type]
            with open(filepath, 'w') as json_file:
                json.dump(self.config_dict, json_file, indent=4)
        else:
            self.console_logger.error(f"Model type '{model_type}' not in supported model types")